from django.db.backends.sqlite3.base import DatabaseWrapper

from io import BytesIO

import boto3
import botocore
import hashlib
import logging
import os


class DatabaseWrapper(DatabaseWrapper):
    """
    Wraps the normal Django SQLite DB engine in an S3 backer!

    """

    def load_remote_db(self):
        """
        Load remote S3 DB
        """

        signature_version = self.settings_dict.get("SIGNATURE_VERSION", "s3v4")
        s3 = boto3.resource(
            's3',
            config=botocore.client.Config(signature_version=signature_version),
        )

        if '/tmp/' not in self.settings_dict['NAME']:
            try:
                etag = ''
                if os.path.isfile('/tmp/' + self.settings_dict['NAME']):
                    m = hashlib.md5()
                    with open('/tmp/' + self.settings_dict['NAME'], 'rb') as f:
                        m.update(f.read())

                    # In general the ETag is the md5 of the file, in some cases it's not,
                    # and in that case we will just need to reload the file, I don't see any other way
                    etag = m.hexdigest()

                obj = s3.Object(self.settings_dict['BUCKET'], self.settings_dict['NAME'])
                obj_bytes = obj.get(IfNoneMatch=etag)["Body"]  # Will throw E on 304 or 404

                with open('/tmp/' + self.settings_dict['NAME'], 'wb') as f:
                    f.write(obj_bytes.read())

                m = hashlib.md5()
                with open('/tmp/' + self.settings_dict['NAME'], 'rb') as f:
                    m.update(f.read())

                self.db_hash = m.hexdigest()

            except botocore.exceptions.ClientError as e:
                if e.response['Error']['Code'] == "304":
                    logging.debug("ETag matches md5 of local copy, using local copy of DB!")
                    self.db_hash = etag
                else:
                    logging.debug("Couldn't load remote DB object.")
            except Exception as e:
                # Weird one
                logging.debug(e)

        # SQLite DatabaseWrapper will treat our tmp as normal now
        # Check because Django likes to call this function a lot more than it should
        if '/tmp/' not in self.settings_dict['NAME']:
            self.settings_dict['REMOTE_NAME'] = self.settings_dict['NAME']
            self.settings_dict['NAME'] = '/tmp/' + self.settings_dict['NAME']

        # Make sure it exists if it doesn't yet
        if not os.path.isfile(self.settings_dict['NAME']):
            open(self.settings_dict['NAME'], 'a').close()

        logging.debug("Loaded remote DB!")

    def __init__(self, *args, **kwargs):
        super(DatabaseWrapper, self).__init__(*args, **kwargs)
        self.db_hash = None
        self.load_remote_db()

    def close(self, *args, **kwargs):
        """
        Engine closed, copy file to DB if it has changed
        """
        super(DatabaseWrapper, self).close(*args, **kwargs)

        signature_version = self.settings_dict.get("SIGNATURE_VERSION", "s3v4")
        s3 = boto3.resource(
            's3',
            config=botocore.client.Config(signature_version=signature_version),
        )

        try:
            with open(self.settings_dict['NAME'], 'rb') as f:
                fb = f.read()

                m = hashlib.md5()
                m.update(fb)
                if self.db_hash == m.hexdigest():
                    logging.debug("Database unchanged, not saving to remote DB!")
                    return

                bytesIO = BytesIO()
                bytesIO.write(fb)
                bytesIO.seek(0)

                s3_object = s3.Object(self.settings_dict['BUCKET'], self.settings_dict['REMOTE_NAME'])
                result = s3_object.put('rb', Body=bytesIO)
        except Exception as e:
            logging.debug(e)

        logging.debug("Saved to remote DB!")



from psycopg2 import connect
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
from django.core.management.base import BaseCommand
from django.conf import settings


class Command(BaseCommand):
    help = 'Creates the initial Postgres database'

    def handle(self, *args, **options):
        self.stdout.write(self.style.SUCCESS('Starting DB creation..'))

        dbname = settings.DATABASES['default']['NAME']
        user = settings.DATABASES['default']['USER']
        password = settings.DATABASES['default']['PASSWORD']
        host = settings.DATABASES['default']['HOST']

        self.stdout.write(self.style.SUCCESS('Connecting to host..'))
        con = connect(dbname='postgres', user=user, host=host, password=password)
        con.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)

        self.stdout.write(self.style.SUCCESS('Creating database'))
        cur = con.cursor()
        cur.execute('CREATE DATABASE ' + dbname)
        cur.close()

        con.close()

        self.stdout.write(self.style.SUCCESS('All done!'))



from psycopg2 import connect
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
from django.core.management.base import BaseCommand
from django.conf import settings


class Command(BaseCommand):
    """
    Command to drop a schema from a PostgreSQL database.
    """
    help = 'Drops a schema from the PostgreSQL database.'

    def handle(self, *args, **options):
        self.stdout.write(self.style.SUCCESS('Starting schema deletion...'))

        dbname = settings.DATABASES['default']['NAME']
        user = settings.DATABASES['default']['USER']
        password = settings.DATABASES['default']['PASSWORD']
        host = settings.DATABASES['default']['HOST']

        con = connect(dbname=dbname, user=user, host=host, password=password)

        self.stdout.write(self.style.SUCCESS('Removing schema {schema} from database {dbname}'
                                             .format(schema=settings.SCHEMA, dbname=dbname)))
        con.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)

        cur = con.cursor()
        cur.execute('DROP SCHEMA {schema} CASCADE;'.format(schema=settings.SCHEMA))
        cur.close()

        con.close()

        self.stdout.write(self.style.SUCCESS('All done.'))



from django.core.management.base import BaseCommand
from django.contrib.auth import get_user_model

import random
import string


class Command(BaseCommand):
    """
    This command will create a default Django admin superuser.
    """
    help = 'Creates a Django admin superuser.'

    def add_arguments(self, parser):
        parser.add_argument('arguments', nargs='*')

    def handle(self, *args, **options):
        # Gets the model for the current Django project's user.
        # This handles custom user models as well as Django's default.
        User = get_user_model()

        self.stdout.write(self.style.SUCCESS('Creating a new admin superuser...'))

        # If the command args are given -> try to create user with given args
        if options['arguments']:
            try:
                user = User.objects.create_superuser(*options['arguments'])
                self.stdout.write(
                    self.style.SUCCESS(
                        'Created the admin superuser "{user}" with the given parameters.'.format(
                            user=user,
                        )
                    )
                )
            except Exception as e:
                self.stdout.write('ERROR: Django returned an error when creating the admin superuser:')
                self.stdout.write(str(e))
                self.stdout.write('')
                self.stdout.write('The arguments expected by the command are in this order:')
                self.stdout.write(str(User.objects.create_superuser.__code__.co_varnames[1:-1]))

        # or create default admin user
        else:
            pw = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(10))
            User.objects.create_superuser('admin', 'admin@admin.com', pw)
            self.stdout.write(self.style.SUCCESS('Created user "admin", email: "admin@admin.com", password: ' + pw))
            self.stdout.write(self.style.SUCCESS('Log in and change this password immediately!'))



from django.core.management.base import BaseCommand
from django.db import connection, DatabaseError
from django.conf import settings


class Command(BaseCommand):
    """
    This command performs a vacuum on the S3SQLite database.
    It is good to do this occasionally to keep the SQLite database stored
    on S3 as small and unfragmented as possible. It is recommended to be
    run after deleting data.
    """
    help = 'Performs a vacuum command on a S3 stored SQLite database to minimize size and fragmentation.'

    def handle(self, *args, **options):
        if settings.DATABASES['default']['ENGINE'] != "zappa_django_utils.db.backends.s3sqlite":
            raise DatabaseError('This command is only for the s3sqlite Django DB engine.')
        else:
            self.stdout.write(self.style.SUCCESS('Starting database VACUUM...'))
            cursor = connection.cursor()
            cursor.execute('VACUUM;')
            cursor.close()
            self.stdout.write(self.style.SUCCESS('VACUUM complete.'))



import os
import sys
from setuptools import setup
from io import open

# New PyPI handles markdown! Yay!
with open('README.md') as f:
    README = f.read()

# Set external files
with open(os.path.join(os.path.dirname(__file__), 'requirements.txt')) as f:
    required = f.read().splitlines()

with open(os.path.join(os.path.dirname(__file__), 'test_requirements.txt')) as f:
    test_required = f.read().splitlines()

setup(
    name='zappa-django-utils',
    version='0.4.0',
    packages=['zappa_django_utils'],
    install_requires=required,
    tests_require=test_required,
    test_suite='nose.collector',
    include_package_data=True,
    license='MIT License',
    description='Helpers for Django Zappa deployments',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/Miserlou/zappa-django-utils',
    author='Rich Jones',
    author_email='rich@openwatch.net',
    classifiers=[
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)
